/******************************************************************************
 *
 * Project:  OSGeo4W
 * Purpose:  Update DLLs in windows\system32 if we have newer version.
 * Author:   Frank Warmerdam, warmerdam@pobox.com
 *
 ******************************************************************************
 * Copyright (c) 2009, Frank Warmerdam
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 ****************************************************************************/

#include <windows.h>
#include <stdio.h>

static void GetDLLVersion( char *pszDLLPath, 
                           __int64 &version, 
                           bool &bExists, bool &bHasVersionInfo );
static void Usage();
                          
/************************************************************************/
/*                                main()                                */
/************************************************************************/

int main( int argc, char **argv )
{
    bool bQuiet = false;
    bool bOnlyIfTargetExists = false;
    bool bDoCopy = false;
    bool bNotifyReboot = false;
    char *pszSourceDLL = NULL;
    char *pszTargetDir = NULL;
    char szTargetDLL[2048] = "";
    int  i;

/* -------------------------------------------------------------------- */
/*      Process arguments.                                              */
/* -------------------------------------------------------------------- */
    if( argc < 2 )
        Usage();

    for( i = 1; i < argc; i++ )
    {
        if( strcmp(argv[i],"-q") == 0 )
            bQuiet = true;
        
        else if( strcmp(argv[i],"-oite") == 0 )
            bOnlyIfTargetExists = true;

        else if( strcmp(argv[i],"-copy") == 0 )
            bDoCopy = true;

        else if( strcmp(argv[i],"-reboot") == 0 )
            bNotifyReboot = true;

        else if( argv[i][0] == '-' )
            Usage();

        else if( pszSourceDLL == NULL )
            pszSourceDLL = argv[i];
        
        else if( pszTargetDir == NULL )
            pszTargetDir = argv[i];

        else
            Usage();
    }

    if( pszSourceDLL == NULL )
        Usage();

/* -------------------------------------------------------------------- */
/*      Use default target directory if one is not specified.           */
/* -------------------------------------------------------------------- */
    if( pszTargetDir == NULL )
    {
        if( getenv("WINDIR") != NULL )
        {
            static char szSystem32[2048];

            sprintf( szSystem32, "%s\\system32", getenv("WINDIR") );
            pszTargetDir = szSystem32;
        }
        else
            pszTargetDir = "C:\\windows\\system32";
    }

/* -------------------------------------------------------------------- */
/*      Get version info on source DLL.                                 */
/* -------------------------------------------------------------------- */
    __int64 nSourceVersion = 0;
    bool    bExists, bHasVersionInfo;

    GetDLLVersion( pszSourceDLL, nSourceVersion, bExists, bHasVersionInfo );

    if( !bExists )
    {
        fprintf( stderr, "%s does not exist.\n", pszSourceDLL );
        exit( 1 );
    }

    if( !bHasVersionInfo )
    {
        fprintf( stderr, "%s has no version info.\n", pszSourceDLL );
        exit( 1 );
    }

    if( !bQuiet )
        printf( "%s: Version=%d.%d.%d.%d\n",
                pszSourceDLL, 
                (int) ((nSourceVersion >> 48) & 0xffff),
                (int) ((nSourceVersion >> 32) & 0xffff),
                (int) ((nSourceVersion >> 16) & 0xffff),
                (int) ((nSourceVersion >> 0) & 0xffff) );

/* -------------------------------------------------------------------- */
/*      Construct the name of the destination DLL.                      */
/* -------------------------------------------------------------------- */
    const char *pszBase;

    for( i = strlen(pszSourceDLL)-1; i > 0; i-- )
    {
        if( pszSourceDLL[i] == '/' 
            || pszSourceDLL[i] == '\\'
            || pszSourceDLL[i] == ':' )
        {
            i++;
            break;
        }
    }
    
    sprintf( szTargetDLL, "%s\\%s", pszTargetDir, pszSourceDLL+i );

/* -------------------------------------------------------------------- */
/*      Get info on the target DLL.                                     */
/* -------------------------------------------------------------------- */
    __int64 nTargetVersion = 0; 

    GetDLLVersion( szTargetDLL, nTargetVersion, bExists, bHasVersionInfo );

    if( !bQuiet )
        printf( "%s: Version=%d.%d.%d.%d\n",
                szTargetDLL,
                (int) ((nTargetVersion >> 48) & 0xffff),
                (int) ((nTargetVersion >> 32) & 0xffff),
                (int) ((nTargetVersion >> 16) & 0xffff),
                (int) ((nTargetVersion >> 0) & 0xffff) );

    if( !bDoCopy )
        exit( 0 );

/* -------------------------------------------------------------------- */
/*      Do we want to update?                                           */
/* -------------------------------------------------------------------- */
    if( !bExists && bOnlyIfTargetExists )
    {
        if( !bQuiet )
            printf( "Target does not exist, no action.\n" );
        exit( 0 );
    }

    if( nTargetVersion >= nSourceVersion )
    {
        if( !bQuiet )
            printf( "Target is not older than source, no action.\n" );
        exit( 0 );
    }

/* -------------------------------------------------------------------- */
/*      Try to copy the file directly.                                  */
/* -------------------------------------------------------------------- */
    if( CopyFile( pszSourceDLL, szTargetDLL, FALSE ) )
    {
        if( !bQuiet )
            printf( "Copied %s to %s successfully.\n",
                    pszSourceDLL, szTargetDLL );
        exit( 0 );
    }

/* -------------------------------------------------------------------- */
/*      If that failed, we are presumably replacing an "in use" DLL.    */
/*      Copy to a .new version and schedule for renaming on reboot.     */
/* -------------------------------------------------------------------- */
    char szTempTargetDLL[2048];

    sprintf( szTempTargetDLL, "%s.new", szTargetDLL );

    if( CopyFile( pszSourceDLL, szTempTargetDLL, FALSE ) )
    {
        if (!MoveFileEx( szTempTargetDLL, szTargetDLL, 
                         MOVEFILE_DELAY_UNTIL_REBOOT |
                         MOVEFILE_REPLACE_EXISTING))
        {
            fprintf( stderr, 
                     "Target file %s is busy/unwritable, and attempt to\n"
                     "schedule %s to be renamed on reboot have failed.\n",
                     szTargetDLL, szTempTargetDLL );
            exit( 1 );
        }

        if( !bQuiet )
        {
            printf( "Target file %s is busy/unwritable, file written to\n"
                    "%s, and will be renamed on reboot.\n",
                    szTargetDLL, szTempTargetDLL );
        }

        if( bNotifyReboot )
        {
            char *pszDir = getenv("OSGEO4W_ROOT");
            if( pszDir )
            {
                char szEtcReboot[2048];
                FILE *fp;

                sprintf(szEtcReboot, "%s\\etc\\reboot", pszDir);
                fp = fopen(szEtcReboot, "w");
                if( !fp )
                {
                    fprintf( stderr, "Could not create file %s\n", szEtcReboot );
                    exit( 1 );
                }

                fclose(fp);
            }
            else
            {
                fprintf( stderr, "Environment variable OSGEO4W_ROOT not set\n" );
                exit(1);
            }

            if( !bQuiet )
            {
                printf( "Installer was notified that a reboot is due.\n" );
            }

        }
        exit( 0 );
    }

    fprintf( stderr, 
             "Target file %s is busy/unwritable, and attempt to copy\n"
             "to temporary file %s has failed.\n",
             szTargetDLL, szTempTargetDLL );

    exit( 1 );
}

/************************************************************************/
/*                               Usage()                                */
/************************************************************************/

static void Usage()

{
    printf( "\n"
            "Usage: dllupdate.exe [-q] [-oite] [-copy] source_dll [target_directory]\n"
            "\n"
            " -q: quiet\n"
            " -oite: only copy if target dll exists\n"
            " -copy: enable copy/update (default is info only)\n" 
            " -reboot: create %%OSGEO4W_ROOT%%/etc/reboot if the update needs a reboot\n" 
            " source_dll: The full path to the source dll\n"
            " target_directory: path to target dir, default is %%WINDIR%%\\system32\n" );
    exit( 1 );
}

/************************************************************************/
/*                           GetDLLVersion()                            */
/************************************************************************/
    
static void GetDLLVersion( char *pszDLLPath, 
                           __int64 &version, 
                           bool &bExists, bool &bHasVersionInfo )

{
    DWORD dwLen, dwUseless, dwBufSize;
    LPTSTR lpVI;
    VS_FIXEDFILEINFO* lpFFI;
    FILE *fp;

    bExists = false;
    bHasVersionInfo = false;

    if( (fp = fopen(pszDLLPath,"rb")) == NULL )
        return;

    bExists = true;
    fclose( fp );

    dwLen = GetFileVersionInfoSize( pszDLLPath, &dwUseless );

    if( dwLen == 0 )
        return;

    lpVI = (LPSTR) malloc(dwLen);
    if( !GetFileVersionInfo( pszDLLPath, NULL, dwLen, lpVI ) )
        return;

    if( !VerQueryValue( lpVI, "\\", (LPVOID *) &lpFFI, (UINT *) &dwUseless ) )
        return;

    bHasVersionInfo = true;

    version = lpFFI->dwProductVersionMS;
    version = version << 32;
    version |= lpFFI->dwProductVersionLS;

    free( lpVI );
}
